/* Exercise redc1.h.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of Redc1.

   Redc1 is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   Redc1 is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with Redc1; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */


#define WANT_ASSERT  1

#include <stdio.h>
#include <stdlib.h>
#include "redc1.h"

#define REDC1_MODULUS_ONE_BIT_SPARE 0

unsigned long tests = 0;

void
check_REDC_IN_OUT (void)
{
  mp_limb_t  N, Ninv, RmodN, R2modN, x, x2, xR, r2, y, yR, z, zR, z2;
  int  i, j;

  for (i = 0; i < 100; i++)
    {
      mpn_random (&N, 1);
      N |= 1;
#if REDC1_MODULUS_ONE_BIT_SPARE
      N &= (MP_LIMB_T_MAX >> 1);
#endif

      REDC1_INIT_Ninv (Ninv, N);
      if (N*Ninv != 1)
        {
          printf ("N*Ninv != 1\n");
          printf ("  N      %#lX\n", N);
          printf ("  Ninv   %#lX\n", Ninv);
          printf ("  N*Ninv %#lX\n", N*Ninv);
          abort ();
        }

      REDC1_INIT_RmodN (RmodN, N);
      r2 = (1ULL << GMP_LIMB_BITS) % N;
      if (RmodN != r2)
        {
          printf ("REDC1_INIT_RmodN wrong\n");
          abort ();
        }

      REDC1_INIT_R2modN (R2modN, N, RmodN);
      r2 = ((unsigned long long) r2 << GMP_LIMB_BITS) % N;
      if (R2modN != r2)
        {
          printf ("REDC1_INIT_R2modN wrong\n");
          printf ("  N           %#lX\n", N);
          printf ("  RmodN       %#lX\n", RmodN);
          printf ("  got R2modN  %#lX\n", R2modN);
          printf ("  want R2modN %#lX\n", r2);
          abort ();
        }

      /* 0 */
      {
        mp_limb_t zero = 0;
        mp_limb_t zeroR_b, zeroR_c;
        REDC1_IN (zeroR_b, zero, N, Ninv, R2modN);
        REDC1_IN_SINGLE (zeroR_c, zero, N);
        if (! (zeroR_b == 0 && zeroR_c == 0))
          {
            printf ("Zero residue wrong\n");
            printf ("  IN        %#lX\n", zeroR_b);
            printf ("  IN_SINGLE %#lX\n", zeroR_c);
            abort ();
          }
        tests++;
      }

      /* 1 */
      {
        mp_limb_t one = 1;
        mp_limb_t oneR_a, oneR_b, oneR_c;
        REDC1_IN_1 (oneR_a, RmodN);
        REDC1_IN (oneR_b, one, N, Ninv, R2modN);
        REDC1_IN_SINGLE (oneR_c, one, N);
        if (! (oneR_a == oneR_b && oneR_a == oneR_c))
          {
            printf ("One residue wrong\n");
            printf ("  IN_1      %#lX\n", oneR_a);
            printf ("  IN        %#lX\n", oneR_b);
            printf ("  IN_SINGLE %#lX\n", oneR_c);
            abort ();
          }
        tests++;
      }

      /* -1 */
      {
        mp_limb_t negone = N-1;
        mp_limb_t negoneR_a, negoneR_b, negoneR_c;
        REDC1_IN_NEG1 (negoneR_a, N, RmodN);
        REDC1_IN (negoneR_b, negone, N, Ninv, R2modN);
        REDC1_IN_SINGLE (negoneR_c, negone, N);
        if (! (negoneR_a == negoneR_b && negoneR_a == negoneR_c))
          {
            printf ("Negone residue wrong\n");
            printf ("  IN_1      %#lX\n", negoneR_a);
            printf ("  IN        %#lX\n", negoneR_b);
            printf ("  IN_SINGLE %#lX\n", negoneR_c);
            abort ();
          }
        tests++;
      }

      for (j = 0; j < 100; j++)
        {
          mpn_random (&x, 1);
          mpn_random (&y, 1);

          REDC1_IN (xR, x, N, Ninv, R2modN);
          r2 = ((unsigned long long) x << GMP_LIMB_BITS) % N;
          if (xR != r2)
            {
              printf ("REDC1_IN wrong\n");
              printf ("  N       %#lX\n", N);
              printf ("  x       %#lX\n", x);
              printf ("  xR      %#lX\n", xR);
              printf ("  want xR %#lX\n", r2);
              abort ();
            }
          tests++;

          REDC1_IN_SINGLE (r2, x, N);
          if (xR != r2)
            {
              printf ("REDC1_IN_SINGLE wrong\n");
              printf ("  N       %#lX\n", N);
              printf ("  R2modN  %#lX\n", R2modN);
              printf ("  x       %#lX\n", x);
              printf ("  xR      %#lX\n", xR);
              printf ("  want xR %#lX\n", r2);
              abort ();
            }
          tests++;

          REDC1_OUT (x2, xR, N, Ninv);
          if (x2 != x % N)
            {
              printf ("REDC1_OUT doesn't give original\n");
              printf ("  N      %#lX\n", N);
              printf ("  Ninv   %#lX\n", Ninv);
              printf ("  x      %#lX\n", x);
              printf ("  xR     %#lX\n", xR);
              printf ("  x out  %#lX\n", x2);
              abort ();
            }
          tests++;

          REDC1_IN (yR, y, N, Ninv, R2modN);
          REDC1_MUL (zR, xR, yR, N, Ninv);
          REDC1_OUT (z, zR, N, Ninv);
          z2 = ((unsigned long long) x * y) % N;
          if (z != z2)
            {
              printf ("REDC1_MUL wrong\n");
            binary:
              printf ("  N      %#lX\n", N);
              printf ("  Ninv   %#lX\n", Ninv);
              printf ("  x      %#lX\n", x);
              printf ("  xR     %#lX\n", xR);
              printf ("  y      %#lX\n", y);
              printf ("  yR     %#lX\n", yR);
              printf ("  zR     %#lX\n", zR);
              printf ("  got z  %#lX\n", z);
              printf ("  want z %#lX\n", z2);
              abort ();
            }
          tests++;

          REDC1_ADD (zR, xR, yR, N);
          REDC1_OUT (z, zR, N, Ninv);
          z2 = ((unsigned long long) x + y) % N;
          if (z != z2)
            {
              printf ("REDC1_ADD wrong\n");
              goto binary;
            }
          tests++;

          REDC1_SUB (zR, xR, yR, N);
          REDC1_OUT (z, zR, N, Ninv);
          z2 = ((unsigned long long) x + N - (y % N)) % N;
          if (z != z2)
            {
              printf ("REDC1_SUB wrong\n");
              goto binary;
            }
          tests++;
        }
    }
}


int
main (void)
{
  setbuf (stdout, NULL);
  setbuf (stderr, NULL);

  check_REDC_IN_OUT ();

  printf ("%lu tests ok\n", tests);
  exit (0);
}
